import functools
import pathlib
from typing import Final, Optional

import ase
from ase.calculators import calculator
import jax
import jaxtyping as jt
import jraph
import numpy as np
import omegaconf
import reax
from tensorial import gcnn
import tensorial.config
from tensorial.gcnn import atomic
from typing_extensions import override

from .. import typing
from ..mlip import keys


class Calculator(calculator.Calculator):
    """
    E3MD's ASE calculator.
    """

    def __init__(
        self,
        model: typing.Model,
        params: jt.PyTree,
        r_max: float,
        accelerator: str = "auto",
        jit=True,
    ):
        super().__init__()
        self.implemented_properties = ["energy", "forces", "stress"]

        # Params
        self._model: Final[typing.Model] = model
        self._params: Final[jt.PyTree] = params
        self._r_max: Final[float] = r_max

        if jit:
            calc = functools.partial(jax.jit(self._model.apply), self._params)
        else:
            calc = functools.partial(self._model.apply, self._params)
        self._calculate = calc
        self._device: jax.Device = (
            jax.devices()[0] if accelerator == "auto" else jax.devices(accelerator)[0]
        )

        # State
        self._graph_padding: Optional[gcnn.data.GraphPadding] = None

    @classmethod
    def from_checkpoint(
        cls,
        config_path: str | pathlib.Path,
        checkpoint_path: str | pathlib.Path,
        checkpointing: reax.training.Checkpointing = None,
    ) -> "Calculator":
        cfg = omegaconf.OmegaConf.load(config_path)
        model: typing.Model = tensorial.config.instantiate(cfg["model"]["model"])
        if checkpointing is None:
            checkpointing = reax.training.get_default_checkpointing()

        ckpt = checkpointing.load(checkpoint_path)
        params = ckpt[reax.training._checkpointing.PARAMS]  # pylint: disable=protected-access

        return Calculator(
            model,
            params,
            r_max=cfg["r_max"],
        )

    @property
    def graph_padding(self) -> Optional[gcnn.data.GraphPadding]:
        return self._graph_padding

    @override
    def calculate(
        self,
        atoms: ase.Atoms = None,
        properties=("energy",),
        system_changes=tuple(calculator.all_changes),
    ):
        super().calculate(atoms)

        graph = self._create_graph(atoms)
        graph = jax.device_put(graph, self._device)
        out = self._calculate(graph)

        self.results = {}

        energy = out.globals[keys.predicted(keys.TOTAL_ENERGY)]
        stress = out.globals[keys.predicted(keys.STRESS)]

        if keys.MASK in out.globals:
            energy = energy[out.globals[keys.MASK]]
            stress = stress[out.globals[keys.MASK]]

        energy = energy.array.item()
        stress = np.array(stress)

        forces = out.nodes[keys.predicted(keys.FORCES)]
        if keys.MASK in out.nodes:
            forces = forces[out.nodes[keys.MASK]]
        forces = np.array(forces)

        self.results["energy"] = energy
        self.results["forces"] = forces
        self.results["stress"] = stress

    def preprocess(self, structures: list[ase.Atoms]):
        max_nodes = 0
        max_edges = 0

        for structure in structures:
            graph: jraph.GraphsTuple = atomic.graph_from_ase(
                structure, self._r_max, use_calculator=False
            )
            max_nodes = max(max_nodes, graph.n_node[0].item())
            max_edges = max(max_edges, graph.n_edge[0].item())

        self._graph_padding = gcnn.data.GraphPadding(max_nodes + 1, max_edges, 2)

    def _create_graph(self, atoms: ase.Atoms) -> jraph.GraphsTuple:
        graph = atomic.graph_from_ase(atoms, self._r_max, use_calculator=False)
        if self._graph_padding is not None:
            graph = gcnn.data.pad_with_graphs(graph, *self._graph_padding)

        return graph
