"""Module for loading ase.Atoms objects as graphs"""

import collections.abc
from typing import Any, Final, Optional

import ase.io
import jraph
from tensorial import gcnn

__all__ = ("AseDataLoader",)


class AseDataLoader(collections.abc.Sequence[jraph.GraphsTuple]):
    def __init__(
        self,
        path: str,
        limit: Optional[int] = None,
        read_kwargs: Optional[dict[str, Any]] = None,
        as_graphs: Optional[dict[str, Any]] = None,
    ):

        # Params
        self._filepath: Final[str] = path
        self._limit: Final[Optional[int]] = limit
        self._to_graphs: Final[dict[str, Any]] = as_graphs
        self._read_kwargs: Final[dict[str, Any]] = self._init_kwargs(limit, read_kwargs)

        try:
            loaded: list[ase.Atoms] = ase.io.read(self._filepath, **self._read_kwargs)

            self._data: list[ase.Atoms | jraph.GraphsTuple] = (
                [loaded] if isinstance(loaded, ase.Atoms) else loaded
            )
        except FileNotFoundError:
            raise ValueError(
                f"Could not load ASE structures, the passed path does not exist: {path}"
            ) from None

        if self._to_graphs and len(self) > 0:
            # Check that the parameters they passed are OK by requesting the first structure to be
            # converged
            self[0]  # noqa, pylint: disable=pointless-statement

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, item: int) -> Any:
        entry = self._data[item]
        if self._to_graphs and not isinstance(entry, jraph.GraphsTuple):
            # Lazily convert the first time
            entry = gcnn.atomic.graph_from_ase(entry, **self._to_graphs)
            self._data[item] = entry

        return entry

    @staticmethod
    def _init_kwargs(
        limit: Optional[int], read_kwargs: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        if read_kwargs is None:
            read_kwargs = {"index": f":{limit}" if limit is not None else ":"}
        else:
            if limit is not None:
                read_kwargs["index"] = f":{limit}"
            elif "index" not in read_kwargs:
                read_kwargs["index"] = ":"

        return read_kwargs
