from collections.abc import Callable, Sequence
import functools
import os
from typing import Any, Final, Optional

import ase
import ase.io
import jraph
import reax
from tensorial import gcnn
from typing_extensions import override

__all__ = ("SiliconDataModule",)


# class SiliconDataLoader(Sequence):
#     """A simple dataset consisting of a silicon MD trajectory"""
#
#     _max_padding: mlip.data.GraphPadding = None
#
#     def __init__(
#         self,
#         r_max: float,
#         data_dir: str = "data/",
#         traj_file: str = "sitraj.xyz",
#         train_val_test_split: Sequence[Union[int, float]] = (0.85, 0.05, 0.1),
#         batch_size: int = 64,
#     ) -> None:


class SiliconDataModule(reax.DataModule):
    """A simple dataset consisting of a silicon MD trajectory"""

    _max_padding: gcnn.data.GraphPadding = None

    def __init__(
        self,
        r_max: float,
        data_dir: str = "data/",
        traj_file: str = "sitraj.xyz",
        train_val_test_split: Sequence[int | float] = (0.8, 0.1, 0.1),
        batch_size: int = 64,
    ) -> None:
        """Initialize a `SiliconDataModule`.

        :param traj_file: The data directory. Defaults to `"data/"`.
        :param train_val_test_split: The train, validation and test split.
        :param batch_size: The batch size. Defaults to `64`.
        """
        super().__init__()

        # Params
        self._rmax = r_max
        self._data_dir: Final[str] = data_dir
        self._traj_file: Final[str] = traj_file
        self._train_val_test_split: Final[Sequence[int | float]] = train_val_test_split
        self._batch_size: Final[int] = batch_size

        # State
        self.batch_size_per_device = batch_size
        self.data_train: Optional[reax.data.Dataset] = None
        self.data_val: Optional[reax.data.Dataset] = None
        self.data_test: Optional[reax.data.Dataset] = None

    @override
    def setup(self, stage: "reax.Stage", /) -> None:
        """Load data. Set variables: `self.data_train`, `self.data_val`, `self.data_test`.

        This method is called by REAX before `trainer.fit()`, `trainer.validate()`,
        `trainer.test()`, and `trainer.predict()`, so be careful not to execute things like random
        split twice! Also, it is called after `self.prepare_data()` and there is a barrier in
        between which ensures that all the processes proceed to `self.setup()` once the data is
        prepared and available for use.

        :param stage: The stage to setup. Either `"fit"`, `"validate"`, `"test"`, or `"predict"`.
        Defaults to ``None``.
        """
        # load and split datasets only if not loaded already
        if not self.data_train and not self.data_val and not self.data_test:
            to_graph: Callable[[ase.Atoms], jraph.GraphsTuple] = functools.partial(
                gcnn.atomic.graph_from_ase,
                r_max=self._rmax,
                atom_include_keys=("numbers", "forces"),
                global_include_keys=("energy", "stress"),
            )

            traj_file = os.path.join(self._data_dir, self._traj_file)
            structures = ase.io.read(traj_file, index=":")
            graphs = list(map(to_graph, structures))

            # Calculate the max padding we will need for any of the batches
            self._max_padding = gcnn.data.GraphBatcher.calculate_padding(
                graphs, self._batch_size, with_shuffle=True
            )

            # Split up the graphs into sets
            train, val, test = reax.data.random_split(
                stage.rng, dataset=graphs, lengths=self._train_val_test_split
            )

            self.data_train = train
            self.data_val = val
            self.data_test = test

    @override
    def train_dataloader(self) -> reax.DataLoader[Any]:
        """Create and return the train dataloader.

        :return: The train dataloader.
        """
        if self.data_train is None:
            raise reax.exceptions.MisconfigurationException(
                "Must call setup() before requesting the dataloader"
            )

        return gcnn.data.GraphLoader(
            self.data_train,
            batch_size=self._batch_size,
            padding=self._max_padding,
            pad=True,
        )

    @override
    def val_dataloader(self) -> reax.DataLoader[Any]:
        """Create and return the validation dataloader.

        :return: The validation dataloader.
        """
        if self.data_val is None:
            raise reax.exceptions.MisconfigurationException(
                "Must call setup() before requesting the dataloader"
            )

        return gcnn.data.GraphLoader(
            self.data_val,
            batch_size=self.batch_size_per_device,
            shuffle=False,
            padding=self._max_padding,
            pad=True,
        )

    @override
    def test_dataloader(self) -> reax.DataLoader[Any]:
        """Create and return the test dataloader.

        :return: The test dataloader.
        """
        if self.data_test is None:
            raise reax.exceptions.MisconfigurationException(
                "Must call setup() before requesting the dataloader"
            )

        return gcnn.data.GraphLoader(
            self.data_test,
            batch_size=self.batch_size_per_device,
            shuffle=False,
            padding=self._max_padding,
            pad=True,
        )
